#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import time
import atexit
import dbus

from homekit.exceptions import AccessoryNotFoundError

from . import gatt
from .manufacturer_data import parse_manufacturer_specific

# 0x004c is the Company Identifier code for Apple Inc. (see Chapter 6.4.2.2 of the spec on page 124)
COID_APPLE = 0x004c


logger = logging.getLogger(__name__)


class Device(gatt.Device):
    """
    A subclass of gatt.Device that makes it a little easier to use for
    implementing BLE HAP. This class only contains thin wrappers on top of
    Device, such as for decoding manufacturer data and ensuring services are
    resolved after connecting to a device. All HAP protocol details are handled
    at a higher level.

    The `name` field fetches the name of the device according to BlueZ's
    `Alias` property. (Note that direct use of the BlueZ `Name` field seems to
    be discouraged).

    The `homekit_discover_data` field contains the decoded ManufacturerData
    retrieved from BlueZ.
    """

    def __init__(self, *args, **kwargs):
        gatt.Device.__init__(self, *args, **kwargs, managed=False)

        self.name = self._properties.Get('org.bluez.Device1', 'Alias')
        self.homekit_discovery_data = self.get_homekit_discovery_data()

    def get_homekit_discovery_data(self):
        """
        Retrieve and decode the latest ManufacturerData from BlueZ for a
        HomeKit device
        """
        try:
            mfr_data = self._properties.Get('org.bluez.Device1', 'ManufacturerData')
        except dbus.exceptions.DBusException as e:
            if e.get_dbus_name() == 'org.freedesktop.DBus.Error.InvalidArgs':
                return {}
            raise

        # convert from dbus.Dictionary({dbus.UInt16(...): dbus.Array([dbus.Byte(...),...])}) to a dict(int: bytes)
        mfr_data = dict((int(k), bytes(bytearray(v))) for (k, v) in mfr_data.items())

        if COID_APPLE not in mfr_data:
            return {}

        parsed = parse_manufacturer_specific(mfr_data[COID_APPLE])

        if parsed['type'] != 'HomeKit':
            return {}

        return parsed

    def connect(self):
        """
        Establishes a connection to a BLE device. This function does not return
        until the services exposed by the device are fully resolved.

        Eventually the connection will timeout and an `AccessoryNotFound` error
        will be raised.
        """
        super().connect()

        try:
            if not self.services:
                logger.debug('waiting for services to be resolved')
                for i in range(20):
                    if self.is_services_resolved():
                        break
                    time.sleep(1)
                else:
                    raise AccessoryNotFoundError('Unable to resolve device services + characteristics')

                # This is called automatically when the mainloop is running, but we
                # want to avoid running it and blocking for an indeterminate amount of time.
                logger.debug('enumerating resolved services')
                self.services_resolved()
        except dbus.exceptions.DBusException:
            raise AccessoryNotFoundError('Unable to resolve device services + characteristics')

    def characteristic_read_value_failed(self, characteristic, error):
        logger.debug('read failed: %s %s', characteristic, error)

    def characteristic_write_value_succeeded(self, characteristic):
        logger.debug('write success: %s', characteristic)

    def characteristic_write_value_failed(self, characteristic, error):
        logger.debug('write failed: %s %s', characteristic, error)


class DeviceManager(gatt.DeviceManager):

    discover_callback = None
    Device = Device

    def __init__(self, adapter):
        super().__init__(adapter)

        # save the old power state of the Bluetooth LE adapter to be able to restore it after
        # the program ends
        atexit.register(self.cleanup)
        self.old_powerstate = self.is_adapter_powered
        self.adapter = adapter
        if not self.old_powerstate:
            logger.debug('Powering on adapter "%s"' % adapter)
            self.is_adapter_powered = True

    def make_device(self, mac_address):
        return self.Device(mac_address=mac_address, manager=self)

    def cleanup(self):
        # restore the old power state
        if not self.old_powerstate:
            logger.debug('restoring power state adapter "%s"' % self.adapter)
            self.is_adapter_powered = self.old_powerstate



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .device import DeviceManager


class DiscoveryDeviceManager(DeviceManager):

    """
    Extension to gatt.DeviceManager that checks if a new device is a HomeKit
    Accessory and tries to parse its HomeKit specific data.
    """

    discover_callback = None

    def make_device(self, mac_address):
        """
        Should not need to call this directly when discovering devices.
        It is called by DeviceManager internals - we subclass it to make it
        ignore non-homekit devices.
        """
        device = super().make_device(mac_address=mac_address)
        if not device.homekit_discovery_data:
            return
        self._manage_device(device)
        if self.discover_callback:
            self.discover_callback(device)

    def start_discovery(self, callback=None):
        """
        Tells BlueZ to start notifying us of detected BLE devices.

        If `callback` is provided and the dbus mainloop is running then your
        function will be called every time a new device is discovered.
        """
        self.discover_callback = callback
        return super().start_discovery()

    def devices(self):
        # The standard implementation of devices causes a fresh of the dbus
        # objects which is not what we want at all
        return self._devices.values()



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


class _Categories(object):
    """
    This data is taken from Table 12-3 Accessory Categories on page 254. Values above 19 are reserved.
    Additional categories ( 20-23 pulled from
    https://github.com/abedinpour/HAS/blob/master/src/categories.ts )
    """
    OTHER = 1
    BRIDGE = 2
    FAN = 3
    GARAGE = 4
    LIGHTBULB = 5
    DOOR_LOCK = 6
    OUTLET = 7
    SWITCH = 8
    THERMOSTAT = 9
    SENSOR = 10
    SECURITY_SYSTEM = 11
    DOOR = 12
    WINDOW = 13
    WINDOW_COVERING = 14
    PROGRAMMABLE_SWITCH = 15
    RANGE_EXTENDER = 16
    IP_CAMERA = 17
    VIDEO_DOOR_BELL = 18
    AIR_PURIFIER = 19
    # new categories referenced from https://github.com/abedinpour/HAS/blob/master/src/categories.ts
    HEATER = 20
    AIRCONDITIONER = 21
    HUMIDIFIER = 22
    DEHUMIDIFER = 23

    def __init__(self):
        self._categories = {
            _Categories.OTHER: 'Other',
            _Categories.BRIDGE: 'Bridge',
            _Categories.FAN: 'Fan',
            _Categories.GARAGE: 'Garage',
            _Categories.LIGHTBULB: 'Lightbulb',
            _Categories.DOOR_LOCK: 'Door Lock',
            _Categories.OUTLET: 'Outlet',
            _Categories.SWITCH: 'Switch',
            _Categories.THERMOSTAT: 'Thermostat',
            _Categories.SENSOR: 'Sensor',
            _Categories.SECURITY_SYSTEM: 'Security System',
            _Categories.DOOR: 'Door',
            _Categories.WINDOW: 'Window',
            _Categories.WINDOW_COVERING: 'Window Covering',
            _Categories.PROGRAMMABLE_SWITCH: 'Programmable Switch',
            _Categories.RANGE_EXTENDER: 'Range Extender',
            _Categories.IP_CAMERA: 'IP Camera',
            _Categories.VIDEO_DOOR_BELL: 'Video Door Bell',
            _Categories.AIR_PURIFIER: 'Air Purifier',
            _Categories.HEATER: "Heater",
            _Categories.AIRCONDITIONER: "Air Conditioner",
            _Categories.HUMIDIFIER: 'Humidifier',
            _Categories.DEHUMIDIFER: "Dehumidifier"
        }

        self._categories_rev = {self._categories[k]: k for k in self._categories.keys()}

    def __contains__(self, item):
        if item in self._categories:
            return True

        if item in self._categories_rev:
            return True

        return False

    def __getitem__(self, item):
        if item in self._categories:
            return self._categories[item]

        if item in self._categories_rev:
            return self._categories_rev[item]

        raise KeyError('Item {item} not found'.format(item=item))


Categories = _Categories()



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import sys
import logging

from homekit.controller import Controller
from homekit.log_support import setup_logging, add_log_arguments


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit ')
    parser.add_argument('-f', action='store', required=True, dest='file',
                        help='HomeKit pairing data file')
    parser.add_argument('-a', action='store', required=True, dest='alias',
                        help='alias for the pairing')
    parser.add_argument('-c', action='store', required=True, dest='connection',
                        choices=['BLE', 'IP'],
                        help='connection type for the pairing')
    parser.add_argument('-k', action='store', required=True, dest='key',
                        help='long term public key for the pairing')
    parser.add_argument('-i', action='store', required=True, dest='id',
                        help='accessory ID for the pairing')
    parser.add_argument('-m', action='store', required=False, dest='mac',
                        help='accessory MAC for the pairing')
    add_log_arguments(parser)
    return parser.parse_args()


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller()
    try:
        controller.load_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    try:
        pairings = controller.get_pairings()
        if args.alias not in pairings:
            print('"{a}" is no known alias'.format(a=args.alias))
            sys.exit(-1)

        pairing_data = pairings[args.alias]._get_pairing_data()
        pairing_data['Connection'] = args.connection
        if args.connection == 'IP':
            pairing_data['AccessoryPairingID'] = args.id
            pairing_data['AccessoryLTPK'] = args.key
        elif args.connection == 'BLE':
            pairing_data['AccessoryPairingID'] = args.id
            pairing_data['AccessoryLTPK'] = args.key
            pairing_data['AccessoryMAC'] = args.mac

        controller.save_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    try:
        pairing = controller.get_pairings()[args.alias]
        data = pairing.list_accessories_and_characteristics()
        controller.save_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging

from homekit.model import Categories
from homekit.model.status_flags import BleStatusFlags


def parse_manufacturer_specific(input_data):
    """
    Parse the manufacturer specific data as returned via Bluez ManufacturerData. This skips the data for LEN, ADT and
    CoID as specified in Chapter 6.4.2.2 of the spec on page 124. Data therefore starts at TY (must be 0x06).

    :param input_data: manufacturer specific data as bytes
    :return: a dict containing the type (key 'type', value 'HomeKit'), the status flag (key 'sf'), human readable
             version of the status flag (key 'flags'), the device id (key 'device_id'), the accessory category
             identifier (key 'acid'), human readable version of the category (key 'category'), the global state number
             (key 'gsn'), the configuration number (key 'cn') and the compatible version (key 'cv')
    """
    logging.debug('manufacturer specific data: %s', input_data.hex())

    # the type must be 0x06 as defined on page 124 table 6-29
    ty = input_data[0]
    input_data = input_data[1:]
    if ty == 0x06:
        ty = 'HomeKit'

        ail = input_data[0]
        logging.debug('advertising interval %s', '{0:02x}'.format(ail))
        length = ail & 0b00011111
        if length != 13:
            logging.debug('error with length of manufacturer data')
        input_data = input_data[1:]

        sf = input_data[0]
        flags = BleStatusFlags[sf]
        input_data = input_data[1:]

        device_id = (':'.join(input_data[:6].hex()[0 + i:2 + i] for i in range(0, 12, 2))).upper()
        input_data = input_data[6:]

        acid = int.from_bytes(input_data[:2], byteorder='little')
        input_data = input_data[2:]

        gsn = int.from_bytes(input_data[:2], byteorder='little')
        input_data = input_data[2:]

        cn = input_data[0]
        input_data = input_data[1:]

        cv = input_data[0]
        input_data = input_data[1:]
        if len(input_data) > 0:
            logging.debug('remaining data: %s', input_data.hex())
        return {'manufacturer': 'apple', 'type': ty, 'sf': sf, 'flags': flags, 'device_id': device_id, 'acid': acid,
                'gsn': gsn, 'cn': cn, 'cv': cv, 'category': Categories[int(acid)]}

    return {'manufacturer': 'apple', 'type': ty}



#!/usr/bin/env python3

#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json
import sys
import argparse
import logging

from homekit.controller import Controller
from homekit.model.characteristics import CharacteristicsTypes
from homekit.model.services import ServicesTypes
from homekit.log_support import setup_logging, add_log_arguments


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit get accessories app')
    parser.add_argument('-f', action='store', required=True, dest='file', help='File with the pairing data')
    parser.add_argument('-a', action='store', required=True, dest='alias', help='alias for the pairing')
    parser.add_argument('-o', action='store', dest='output', default='compact', choices=['json', 'compact'],
                        help='Specify output format')
    parser.add_argument('--adapter', action='store', dest='adapter', default='hci0',
                        help='the bluetooth adapter to be used (defaults to hci0)')
    add_log_arguments(parser)
    return parser.parse_args()


if __name__ == '__main__':
    args = setup_args_parser()

    setup_logging(args.loglevel)

    controller = Controller(args.adapter)
    try:
        controller.load_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    if args.alias not in controller.get_pairings():
        print('"{a}" is no known alias'.format(a=args.alias))
        sys.exit(-1)

    try:
        pairing = controller.get_pairings()[args.alias]
        data = pairing.list_accessories_and_characteristics()
        controller.save_data(args.file)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    # prepare output
    if args.output == 'json':
        print(json.dumps(data, indent=4))

    if args.output == 'compact':
        for accessory in data:
            aid = accessory['aid']
            for service in accessory['services']:
                s_type = service['type']
                s_iid = service['iid']
                print('{aid}.{iid}: >{stype}<'.format(aid=aid, iid=s_iid, stype=ServicesTypes.get_short(s_type)))

                for characteristic in service['characteristics']:
                    c_iid = characteristic['iid']
                    value = characteristic.get('value', '')
                    c_type = characteristic['type']
                    perms = ','.join(characteristic['perms'])
                    desc = characteristic.get('description', '')
                    c_type = CharacteristicsTypes.get_short(c_type)
                    print('  {aid}.{iid}: {value} ({description}) >{ctype}< [{perms}]'.format(aid=aid,
                                                                                              iid=c_iid,
                                                                                              value=value,
                                                                                              ctype=c_type,
                                                                                              perms=perms,
                                                                                              description=desc))



#!/usr/bin/env python3
#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import sys
import logging

from homekit.log_support import setup_logging, add_log_arguments
from homekit.controller import Controller


def setup_args_parser():
    parser = argparse.ArgumentParser(description='HomeKit BLE discover app - '
                                                 'list all HomeKit Bluetooth LE devices in range')
    parser.add_argument('-t', action='store', required=False, dest='timeout', type=int, default=10,
                        help='Number of seconds to wait (defaults to 10)')
    parser.add_argument('--adapter', action='store', dest='adapter', default='hci0',
                        help='the bluetooth adapter to be used (defaults to hci0)')
    parser.add_argument('-u', action='store_true', required=False, dest='unpaired_only',
                        help='If activated, this option will show only unpaired HomeKit BLE Devices')
    add_log_arguments(parser)
    return parser.parse_args()


if __name__ == '__main__':
    args = setup_args_parser()
    setup_logging(args.loglevel)

    try:
        devices = Controller.discover_ble(args.timeout, args.adapter)
    except Exception as e:
        print(e)
        logging.debug(e, exc_info=True)
        sys.exit(-1)

    print()
    for device in devices:
        if args.unpaired_only and device['sf'] == 0:
            continue
        print('Name: {name}'.format(name=device['name']))
        print('MAC: {mac}'.format(mac=device['mac']))
        print('Configuration number (cn): {conf}'.format(conf=device['cn']))
        print('Device ID (id): {id}'.format(id=device['device_id']))
        print('Compatible Version (cv): {cv}'.format(cv=device['cv']))
        print('Global State Number (s#): {sn}'.format(sn=device['gsn']))
        print('Status Flags (sf): {sf} (Flag: {flags})'.format(sf=device['flags'], flags=device['sf']))
        print('Category Identifier (ci): {c} (Id: {ci})'.format(c=device['category'], ci=device['acid']))
        print()
    sys.exit(0)



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = [
    'Controller', 'BluetoothAdapterError', 'AccessoryDisconnectedError', 'AccessoryNotFoundError',
    'AlreadyPairedError', 'AuthenticationError', 'BackoffError', 'BusyError', 'CharacteristicPermissionError',
    'ConfigLoadingError', 'ConfigSavingError', 'ConfigurationError', 'FormatError', 'HomeKitException',
    'HttpException', 'IncorrectPairingIdError', 'InvalidAuthTagError', 'InvalidError', 'InvalidSignatureError',
    'MaxPeersError', 'MaxTriesError', 'ProtocolError', 'RequestRejected', 'UnavailableError', 'UnknownError',
    'UnpairedError'
]

from homekit.controller import Controller
from homekit.exceptions import BluetoothAdapterError, AccessoryDisconnectedError, AccessoryNotFoundError, \
    AlreadyPairedError, AuthenticationError, BackoffError, BusyError, CharacteristicPermissionError, \
    ConfigLoadingError, ConfigSavingError, ConfigurationError, FormatError, HomeKitException, HttpException, \
    IncorrectPairingIdError, InvalidAuthTagError, InvalidError, InvalidSignatureError, MaxPeersError, MaxTriesError, \
    ProtocolError, RequestRejected, UnavailableError, UnknownError, UnpairedError

from homekit.tools import IP_TRANSPORT_SUPPORTED

if IP_TRANSPORT_SUPPORTED:
    # TODO: change import and let it be imported from its specific file
    from homekit.accessoryserver import AccessoryServer  # noqa: F401

    __all__ += 'AccessoryServer'



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from homekit.model import get_id
from homekit.model.characteristics import OnCharacteristicMixin, BrightnessCharacteristicMixin, \
    HueCharacteristicMixin, SaturationCharacteristicMixin
from homekit.model.services import ServicesTypes, AbstractService


class BHSLightBulbService(AbstractService, OnCharacteristicMixin, BrightnessCharacteristicMixin,
                          HueCharacteristicMixin, SaturationCharacteristicMixin):
    """
    Defined on page 217; Light bulb that supports color selection via Brightness, Hue and Saturation
    """

    def __init__(self):
        AbstractService.__init__(self, ServicesTypes.get_uuid('public.hap.service.lightbulb'), get_id())
        OnCharacteristicMixin.__init__(self, get_id())
        BrightnessCharacteristicMixin.__init__(self, get_id())
        HueCharacteristicMixin.__init__(self, get_id())
        SaturationCharacteristicMixin.__init__(self, get_id())



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This module contains backports of upcoming gatt-python features that havent
been merged yet.

It should only contain code required for these feature and to enable those
features to work.

Right now this module contains:

 * https://github.com/getsenic/gatt-python/pull/36 (GATT Descriptors)
 * https://github.com/getsenic/gatt-python/pull/38 (Discovery timeouts)

This is to allow us to ship homekit_python even though upstream hasn't exposed
all the functionality we require.

Any homekit specific functionality that we don't plan to upstream should subclass
the classes in this module.
"""

import re
import logging
import dbus
import gatt
from gatt.gatt_linux import _error_from_dbus_error
from gi.repository import GObject

from homekit.controller.ble_impl.tools import hci_adapter_exists_and_supports_bluetooth_le, hci_adapter_exists
from homekit.exceptions import BluetoothAdapterError


class Descriptor:

    def __init__(self, characteristic, path, uuid):
        self.characteristic = characteristic
        self.uuid = uuid
        self._bus = characteristic._bus
        self._path = path
        self._object = self._bus.get_object('org.bluez', self._path)

    def read_value(self, offset=0):
        try:
            val = self._object.ReadValue(
                {'offset': dbus.UInt16(offset, variant_level=1)},
                dbus_interface='org.bluez.GattDescriptor1')
            return val
        except dbus.exceptions.DBusException as e:
            error = _error_from_dbus_error(e)
            self.service.device.descriptor_read_value_failed(self, error=error)


class Characteristic(gatt.Characteristic):

    def __init__(self, service, path, uuid):
        gatt.Characteristic.__init__(self, service, path, uuid)

        descriptor_regex = re.compile(self._path + '/desc[0-9abcdef]{4}$')
        self.descriptors = [
            Descriptor(self, desc[0], desc[1]['org.bluez.GattDescriptor1']['UUID'])
            for desc in self._object_manager.GetManagedObjects().items()
            if descriptor_regex.match(desc[0])
        ]


class Service(gatt.Service):

    def characteristics_resolved(self):
        self._disconnect_characteristic_signals()

        characteristics_regex = re.compile(self._path + '/char[0-9abcdef]{4}$')
        managed_characteristics = [
            char for char in self._object_manager.GetManagedObjects().items()
            if characteristics_regex.match(char[0])]
        self.characteristics = [Characteristic(
            service=self,
            path=c[0],
            uuid=c[1]['org.bluez.GattCharacteristic1']['UUID']) for c in managed_characteristics]

        self._connect_characteristic_signals()


class Device(gatt.Device):

    def services_resolved(self):
        self._disconnect_service_signals()

        services_regex = re.compile(self._device_path + '/service[0-9abcdef]{4}$')
        managed_services = [
            service for service in self._object_manager.GetManagedObjects().items()
            if services_regex.match(service[0])]
        self.services = [Service(
            device=self,
            path=service[0],
            uuid=service[1]['org.bluez.GattService1']['UUID']) for service in managed_services]

        self._connect_service_signals()


class DeviceManager(gatt.DeviceManager):

    def __init__(self, adapter_name='hci0'):
        """
        Creates a new DeviceManager but performs a check first to verify the adapter exists and supports
        Bluetooth LE.

        :param adapter_name: the name of the adapter (defaults to hci0)
        :raises BluetoothAdapterError: if either the adapter does not exist or lacks appropriate capabilities
        """
        self.logger = logging.getLogger('homekit.controller.ble')
        if not hci_adapter_exists(adapter_name):
            raise BluetoothAdapterError('Adapter "{a}" does not exist!'.format(a=adapter_name))
        if not hci_adapter_exists_and_supports_bluetooth_le(adapter_name):
            print('Adapter "{a}" seems not to support Bluetooth LE, the command might not function properly!'
                  .format(a=adapter_name))
        super().__init__(adapter_name)

    def set_timeout(self, timeout):
        GObject.timeout_add(timeout, self.stop)

    def make_device(self, mac_address):
        return Device(mac_address=mac_address, manager=self)



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# flake8 is suppressed here because in this file it should just be checked if the packages can be imported.
try:
    import gatt  # noqa: F401
    import dbus  # noqa: F401
    BLE_TRANSPORT_SUPPORTED = True
except ImportError:
    BLE_TRANSPORT_SUPPORTED = False

try:
    import zeroconf  # noqa: F401
    IP_TRANSPORT_SUPPORTED = True
except ImportError:
    IP_TRANSPORT_SUPPORTED = False



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from homekit.model import ToDictMixin


class AbstractService(ToDictMixin):
    def __init__(self, service_type: str, iid: int):
        if type(self) is AbstractService:
            raise Exception('AbstractService is an abstract class and cannot be instantiated directly')
        self.type = service_type
        self.iid = iid
        self.characteristics = []

    def append_characteristic(self, characteristic):
        """
        Append the given characteristic to the service.

        :param characteristic: a subclass of AbstractCharacteristic
        """
        self.characteristics.append(characteristic)

    def to_accessory_and_service_list(self):
        characteristics_list = []
        for c in self.characteristics:
            characteristics_list.append(c.to_accessory_and_service_list())
        d = {
            'iid': self.iid,
            'type': self.type,
            'characteristics': characteristics_list
        }
        return d



#
# Copyright 2018 Joachim Lusiardi
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='homekit',
    packages=setuptools.find_packages(exclude=['tests']),
    version='0.15.0',
    description='Python code to interface HomeKit Accessories and Controllers',
    author='Joachim Lusiardi',
    author_email='pypi@lusiardi.de',
    url='https://github.com/jlusiardi/homekit_python',
    download_url='https://github.com/jlusiardi/homekit_python/archive/0.15.0.tar.gz',
    keywords=['HomeKit'],
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Home Automation',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop'
    ],
    install_requires=[
        'hkdf',
        'ed25519',
        'cryptography>=2.5',
    ],
    extras_require={
        'IP': ['zeroconf'],
        'BLE': ['dbus-python', 'gatt', 'pygobject']
    },
    license='Apache License 2.0',
    long_description=long_description,
    long_description_content_type="text/markdown",
)
